import torch
import cupy as cp
import numpy as np
import math
from .CustomKernel import CustomKernel, Stream
from torchpq.util import get_absolute_path

class GetDivOfAddressCUDA(CustomKernel):
  def __init__(
      self,
      ta=1,
      tpb=256,
      sm_size=48*256*4,
    ):
    self.ta = ta # how many clusters each thread is responsible of
    self.tpb = tpb
    self.sm_size = sm_size

    self._use_torch_in_cupy_malloc()
    self.stream = Stream(torch.cuda.current_stream().cuda_stream)
    with open(get_absolute_path("kernels", "GetDivOfAddressKernel.cu"), "r") as f:
      self.kernel = f.read()
    kernel = (self.kernel
      .replace("_TA_", str(ta))
      .replace("_TPB_", str(tpb))
    )

    self.fn = cp.RawKernel(
      kernel,
      'get_div_of_address',
      backend='nvcc',
      # options=('--maxrregcount=255',),
    )

  def __call__(self, address, div_start, div_end):
    """
      address: [n_data]
      div_start: [n_clusters]
      div_end: [n_clusters]
    """
    n_data = address.shape[0]
    assert div_start.shape[0] == div_end.shape[0]
    n_clusters = div_start.shape[0]
    divs = torch.ones_like(address) * -1

    threads_per_block = (self.tpb,)
    blocks_per_grid = (math.ceil(n_clusters / (self.tpb * self.ta)), )

    # torch.cuda.synchronize()
    self.fn(
      grid=blocks_per_grid,
      block=threads_per_block,
      args=[
        address.data_ptr(),
        div_start.data_ptr(),
        div_end.data_ptr(),
        divs.data_ptr(),
        n_data, n_clusters
        ],
      stream=self.stream
    )
    return divs


import torch
import cupy as cp
import numpy as np
import math
from torchpq.kernels.CustomKernel import CustomKernel, Stream
from torchpq.util import get_absolute_path

class MaxSimCUDA(CustomKernel): 
  def __init__(self, m=None, n=None, k=None, dim=None, distance="euclidean"):
    self.m = m
    self.n = n
    self.k = k
    self.dim = dim
    self.distance = distance
    with open(get_absolute_path("kmeans", "kernels","MaxSimKernel.cu"),'r') as f:
      self.kernel = f.read()

    if distance in ["euclidean", "l2"]:
      distfn = "thread_nseuclidean"
    elif distance in ["manhattan", "l1"]:
      distfn = "thread_nmanhattan"
    elif distance == "inner":
      distfn = "thread_matmul"
    elif distance == "cosine":
      print("warning: input matrices will not be normalized, please normalize them manually for cosine similarity")
      distfn = "thread_matmul"
    else:
      raise ValueError("unrecognized distance type")
      
    self.kernel = (self.kernel
      .replace("_M_", str(m) if m else "M")
      .replace("_N_", str(n) if n else "N")
      .replace("_K_", str(k) if k else "K")
      .replace("_DIM_", str(dim) if dim else "DIM")
      .replace("_DISTFN_", distfn)
    )
    self._use_torch_in_cupy_malloc()
    self.stream = Stream(torch.cuda.current_stream().cuda_stream)
    self._raw_module = cp.RawModule(
      code=self.kernel,
      backend='nvcc',
      options=('--maxrregcount=128', '--use_fast_math'),
    )
    self._fn_tt = cp.RawKernel(
      code=self.kernel,
      name="max_sim_tt",
      backend='nvcc',
      options=('--maxrregcount=128', '--use_fast_math')
    )
    self._fn_nn = cp.RawKernel(
      code=self.kernel,
      name="max_sim_nn",
      backend='nvcc',
      options=('--maxrregcount=128', '--use_fast_math')
    )
    self._fn_tn = cp.RawKernel(
      code=self.kernel,
      name="max_sim_tn",
      backend='nvcc',
      options=('--maxrregcount=128', '--use_fast_math')
    )
    self._fn_nt = cp.RawKernel(
      code=self.kernel,
      name="max_sim_nt",
      backend='nvcc',
      options=('--maxrregcount=128', '--use_fast_math')
    )

  def _call_nn(self, A, B, dim=1):
    """
      Performs the following: 
        vals = max(A @ B, dim);
        inds = argmax(A @ B, dim)
      A: shape = [l, m, k]
      B: shape = [l, k, n]
      dim: 1 or 2. 0 is not supported.
      returns (vals, inds)
      vals: shape = [l, n] if dim=1, [l, m] if dim=2
      inds: shape = [l, n] if dim=1, [l, m] if dim=2
    """
    assert A.shape[0] == B.shape[0]
    assert A.shape[2] == B.shape[1]
    assert dim in (1, 2)
    assert A.device.type == "cuda"
    assert B.device.type == "cuda"
    assert A.dtype in (torch.float, torch.half)
    assert B.dtype in (torch.float, torch.half)
    
    l, m, k = A.shape
    l, k, n = B.shape
    
    if self.m is not None: assert m == self.m
    if self.n is not None: assert n == self.n
    if self.k is not None: assert k == self.k
    if self.dim is not None: assert dim == self.dim


    if dim == 1:
      vals = torch.full([l, n], fill_value=float("-inf"), device="cuda:0", dtype=A.dtype)
      inds = torch.empty([l, n], device="cuda:0", dtype=torch.long)
    elif dim == 2:
      vals = torch.full([l, m], fill_value=float("-inf"), device="cuda:0", dtype=A.dtype)
      inds = torch.empty([l, m], device="cuda:0", dtype=torch.long)

    threads_per_block = (256,)
    blocks_per_grid = (l, math.ceil(n/128), math.ceil(m/128))

    self._fn_nn(
      grid=blocks_per_grid,
      block=threads_per_block,
      args=[
        A.data_ptr(),
        B.data_ptr(),
        vals.data_ptr(),
        inds.data_ptr(),
        m, n, k, dim
      ],
      stream=self.stream
    )
    return (vals, inds)

  def _call_tt(self, A, B, dim=1):
    """
      Performs the following: 
        vals = max(A.T @ B.T, dim);
        inds = argmax(A.T @ B.T, dim)
      A: shape = [l, k, m]
      B: shape = [l, n, k]
      dim: 1 or 2. 0 is not supported.
      returns (vals, inds)
      vals: shape = [l, n] if dim=1, [l, m] if dim=2
      inds: shape = [l, n] if dim=1, [l, m] if dim=2
    """
    assert A.shape[0] == B.shape[0]
    assert A.shape[1] == B.shape[2]
    assert dim in (1, 2)
    assert A.device.type == "cuda"
    assert B.device.type == "cuda"
    assert A.dtype in (torch.float, torch.half)
    assert B.dtype in (torch.float, torch.half)
    
    l, k, m = A.shape
    l, n, k = B.shape

    if self.m is not None: assert m == self.m
    if self.n is not None: assert n == self.n
    if self.k is not None: assert k == self.k
    if self.dim is not None: assert dim == self.dim


    if dim == 1:
      vals = torch.full([l, n], fill_value=float("-inf"), device="cuda:0", dtype=A.dtype)
      inds = torch.empty([l, n], device="cuda:0", dtype=torch.long)
    elif dim == 2:
      vals = torch.full([l, m], fill_value=float("-inf"), device="cuda:0", dtype=A.dtype)
      inds = torch.empty([l, m], device="cuda:0", dtype=torch.long)


    threads_per_block = (256,)
    blocks_per_grid = (l, math.ceil(n/128), math.ceil(m/128))

    self._fn_tt(
      grid=blocks_per_grid,
      block=threads_per_block,
      args=[
        A.data_ptr(),
        B.data_ptr(),
        vals.data_ptr(),
        inds.data_ptr(),
        m, n, k, dim
      ],
      stream=self.stream
    )
    return (vals, inds)

  def _call_tn(self, A, B, dim=1):
    """
      Performs the following: 
        vals = max(A.T @ B, dim);
        inds = argmax(A.T @ B, dim)
      A: shape = [l, k, m]
      B: shape = [l, k, n]
      dim: 1 or 2. 0 is not supported.
      returns (vals, inds)
      vals: shape = [l, n] if dim=1, [l, m] if dim=2
      inds: shape = [l, n] if dim=1, [l, m] if dim=2
    """
    assert A.shape[0] == B.shape[0]
    assert A.shape[1] == B.shape[1]
    assert dim in (1, 2)
    assert A.device.type == "cuda"
    assert B.device.type == "cuda"
    assert A.dtype in (torch.float, torch.half)
    assert B.dtype in (torch.float, torch.half)

    l, k, m = A.shape
    l, k, n = B.shape

    if self.m is not None: assert m == self.m
    if self.n is not None: assert n == self.n
    if self.k is not None: assert k == self.k
    if self.dim is not None: assert dim == self.dim

    if dim == 1:
      vals = torch.full([l, n], fill_value=float("-inf"), device="cuda:0", dtype=A.dtype)
      inds = torch.empty([l, n], device="cuda:0", dtype=torch.long)
    elif dim == 2:
      vals = torch.full([l, m], fill_value=float("-inf"), device="cuda:0", dtype=A.dtype)
      inds = torch.empty([l, m], device="cuda:0", dtype=torch.long)

    threads_per_block = (256,)
    blocks_per_grid = (l, math.ceil(n/128), math.ceil(m/128))

    self._fn_tn(
      grid=blocks_per_grid,
      block=threads_per_block,
      args=[
        A.data_ptr(),
        B.data_ptr(),
        vals.data_ptr(),
        inds.data_ptr(),
        m, n, k, dim
      ],
      stream=self.stream,
    )
    return (vals, inds)

  def _call_nt(self, A, B, dim=1):
    """
      Performs the following: 
        vals = max(A @ B.T, dim);
        inds = argmax(A @ B.T, dim)
      A: shape = [l, m, k]
      B: shape = [l, n, k]
      dim: 1 or 2; 0 is not supported.
      returns (vals, inds)
      vals: shape = [l, n] if dim=1, [l, m] if dim=2
      inds: shape = [l, n] if dim=1, [l, m] if dim=2
    """
    assert A.shape[0] == B.shape[0]
    assert A.shape[2] == B.shape[2]
    assert dim in (1, 2)
    assert A.device.type == "cuda"
    assert B.device.type == "cuda"
    assert A.dtype in (torch.float, torch.half)
    assert B.dtype in (torch.float, torch.half)

    l, m, k = A.shape
    l, n, k = B.shape

    if self.m is not None: assert m == self.m
    if self.n is not None: assert n == self.n
    if self.k is not None: assert k == self.k
    if self.dim is not None: assert dim == self.dim

    if dim == 1:
      vals = torch.full([l, n], fill_value=float("-inf"), device="cuda:0", dtype=A.dtype)
      inds = torch.empty([l, n], device="cuda:0", dtype=torch.long)
    elif dim == 2:
      vals = torch.full([l, m], fill_value=float("-inf"), device="cuda:0", dtype=A.dtype)
      inds = torch.empty([l, m], device="cuda:0", dtype=torch.long)

    threads_per_block = (256,)
    blocks_per_grid = (l, math.ceil(n/128), math.ceil(m/128))

    self._fn_nt(
      grid=blocks_per_grid,
      block=threads_per_block,
      args=[
        A.data_ptr(),
        B.data_ptr(),
        vals.data_ptr(),
        inds.data_ptr(),
        m, n, k, dim
      ],
      stream=self.stream
    )
    return (vals, inds)

  def __call__(self, A, B, dim=1, mode="nn"):
    """
      Performs the following:
        vals = max(f(A) @ g(B), dim)
        inds = argmax(f(A) @ g(B), dim)
      A: torch.Tensor, shape : [l, m, k] or [l, k, m]
      B: torch.Tensor, shape : [l, n, k] or [l, k, n]
      return: (vals, inds)
      vals: torch.Tensor, shape: [l, m] if dim == 2, [l, n] if dim == 1
      inds: torch.Tensor, shape: [l, m] if dim == 2, [l, n] if dim == 1
      dim: int, default : 1
      mode: str, default: "nn"
      Notes:
        f() and g() are determined by mode
        "nn" --> A @ B
        "tt" --> A.T @ B.T
        "nt" --> A @ B.T
        "tn" --> A.T @ B
        dim should be 1 or 2, 0 is not supported.
    """
    assert len(A.shape) == len(B.shape)
    A = A.contiguous()
    B = B.contiguous()
    if len(A.shape) == 2 and len(B.shape) == 2:
      A2 = A[None]
      B2 = B[None]
      dim += 1
    elif len(A.shape) == 3 and len(B.shape) == 3:
      A2 = A
      B2 = B
    else:
      raise ValueError("shape of A and B need to be 2d or 3d")

    if mode == "nn":
      vals, inds = self._call_nn(A2, B2, dim)
    elif mode == "tt":
      vals, inds = self._call_tt(A2, B2, dim)
    elif mode == "tn":
      vals, inds = self._call_tn(A2, B2, dim)
    elif mode == "nt":
      vals, inds = self._call_nt(A2, B2, dim)

    if len(A.shape) == 2 and len(B.shape) == 2:
      vals, inds = vals[0], inds[0]
    return (vals, inds)


from . import kmeans
from . import kernels

from .IVFPQBase import IVFPQBase
from .PQ import PQ

from .IVFPQ import IVFPQ
from .IVFPQR import IVFPQR

try:
  import cupy
except ModuleNotFoundError:
  raise ModuleNotFoundError("cupy is not installed, please visit https://pypi.org/project/cupy/")


import torch
import cupy as cp
import numpy as np
import math
from .CustomKernel import CustomKernel, Stream
from torchpq.util import get_absolute_path

class GetIOACUDA(CustomKernel):
  def __init__(
      self,
      tpb=256,
      sm_size=48*256*4,
    ):
    self.tpb = tpb
    self.sm_size = sm_size

    self._use_torch_in_cupy_malloc()
    self.stream = Stream(torch.cuda.current_stream().cuda_stream)
    with open(get_absolute_path("kernels", "GetIOAKernel.cu"), "r") as f:
      self.kernel = f.read()
    kernel = (self.kernel
      .replace("_TPB_", str(tpb))
    )

    self.fn = cp.RawKernel(
      kernel,
      'get_ioa',
      backend='nvcc',
      # options=('--maxrregcount=255',),
    )

  def __call__(self, labels, unique_labels=None):
    """
      labels: [n_labels]
      unique_labels: [n_unique]
    """
    n_labels = labels.shape[0]
    if unique_labels is None:
      unique_labels = torch.unique(labels)
    n_unique = unique_labels.shape[0]

    ioa = torch.zeros_like(labels) - 2
    threads_per_block = (self.tpb,)
    blocks_per_grid = (math.ceil(n_unique / self.tpb), )

    # torch.cuda.synchronize()
    self.fn(
      grid=blocks_per_grid,
      block=threads_per_block,
      args=[
        labels.data_ptr(),
        unique_labels.data_ptr(),
        ioa.data_ptr(),
        n_labels, n_unique
        ],
      stream=self.stream
    )
    return ioa


import torch
import numpy as np

from torchpq import IVFPQBase
from torchpq import IVFPQTopk
from torchpq import PQ
from .kmeans import KMeans

class IVFPQ(IVFPQBase):
  def __init__(
    self,
    d_vector,
    n_subvectors=8,
    n_cq_clusters=128,
    n_pq_clusters=256,
    blocksize=64,
    verbose=0,
    distance="euclidean",
    device='cuda:0'
    ):
    """
    this is an efficient implementation of IVFPQ algorithm
    for more details, you can search 'Product Quantization for Nearest Neighbor Search'

    Parameters:
      d_vector: int
        dimentionality of vectors to be quantized.

      n_subvectors: int, default : 8
        number of sub-quantizers, needs to be a multiple of 4,
        maximum possible n_subvectors depends on GPU architecture
          GPU Architecture:   fp32  fp16 
          Ampere (GA100)      163   326  
          Turing (TU102 etc.) 64    128  
          Volta (GV100)       96    192  
          Pascal and before   48    96   
      
      n_cq_clusters: int, default : 128
        number coarse quantizer clusters
        recommended value is between 4*sqrt(n_data) ~ 16*sqrt(n_data)

      n_pq_clusters: int, default : 256
        number of product quantizer clusters
        any value other than 256 is not recommended.

      blocksize: int, default : 64
        number of vectors that can be asigned to each cluster of coarse_quantizer initially
        can be expanded using .expand methods, .add method will automatically cal .expand if necessary
        recommended value is (n_data_to_be_stored / n_cq_clusters)

      verbose: int, default : 0
        verbosity

      distance: str, default : 'euclidean'
        type of distance metric
        can be one of: ['euclidean', 'cosine', 'inner']

      device: str, default : 'cuda:0'
        which device to use
        can be one of: ['cpu', 'cuda', 'cuda:X'] *X represents cuda device index
      
    """
    n_cs = 4
    self.n_cs = n_cs
    assert n_subvectors % self.n_cs == 0, f"n_subvectors needs to be a multiple of {n_cs}"
    self.n_probe = 1
    self.d_vector = d_vector
    self.d_subvector = d_vector // n_subvectors
    self.n_pq_clusters = n_pq_clusters
    
    cc = self.get_cc() #compute capability
    if cc[0] < 7 or cc == (7, 2):
      assert n_subvectors <= 48
    elif cc == (7, 0):
      assert n_subvectors <= 96
    elif cc == (7, 5):
      assert n_subvectors <= 64
    elif cc == (8, 0):
      assert n_subvectors <= 163
    elif cc == (8, 6):
      assert n_subvectors <= 99
    
    super(IVFPQ, self).__init__(
      code_size=n_subvectors,
      n_cq_clusters=n_cq_clusters,
      blocksize=blocksize,
      verbose=verbose,
      distance=distance,
      device=device
    )

    self.coarse_q = KMeans(
      max_iter=25,
      n_clusters=n_cq_clusters,
      distance="inner" if distance == "cosine" else distance,
      init_mode="random",
      verbose=verbose,
    )
    self.product_q = PQ(
      d_vector=d_vector,
      n_subvectors=n_subvectors,
      n_clusters=n_pq_clusters,
      distance="inner" if distance == "cosine" else distance,
      verbose=verbose,
    )

    self._topk_fn = IVFPQTopk(
      n_subvectors=n_subvectors,
      n_clusters=n_pq_clusters,
      n_cs=n_cs
    )

  def set_coarse_q_max_iter(self, value):
    self.coarse_q.max_iter = value

  def set_product_q_max_iter(self, value):
    self.product_q.kmeans.max_iter = value
  
  def set_coarse_q_n_redo(self, value):
    self.coarse_q.n_redo = value

  def set_product_q_n_redo(self, value):
    self.product_q.kmeans.n_redo = value

  def add(self, input, input_ids=None):
    """
      input: torch.Tensor, shape : [n_data, d_vector], dtype : float32
      input_ids: torch.Tensor, shape : [n_data], dtype : int64
    """
    assert self._is_trained == True, "Module is not trained"
    if self.distance == "cosine":
      input = self.normalize(input)
    input = input.contiguous()

    d_vector, n_data = input.shape
    assert d_vector == self.d_vector
    labels = self.coarse_q.predict(input)

    unique_labels, counts = torch.unique(labels, return_counts=True)
    ioa = self.get_ioa(labels, unique_labels)

    # expand storage if necessary
    while True:
      free_space = self.div_capacity[labels] - self.div_size[labels] - (ioa + 1)
      expansion_required = labels[free_space < 0].unique()
      if expansion_required.shape[0] == 0:
        break
      self.expand(expansion_required)

    #get write address
    empty_adr = torch.nonzero(self.is_empty == 1)[:, 0] #[n_empty]
    div_of_empty_adr = self.get_div_of_address(empty_adr) #[n_empty]
    write_address = self.get_write_address(
      empty_adr=empty_adr,
      labels=labels,
      ioa=ioa
    )
    # quantize input and store
    quantized_input = self.product_q.encode(input).byte() #[n_subvector, n_data]
    quantized_input = quantized_input.reshape(
      self.n_subvectors // self.n_cs,
      self.n_cs,
      n_data
    ).transpose(1, 2)
    self.set_data_of_address(quantized_input, write_address)

    #store ids
    if input_ids is None:
      if self.tot_size == 0:
        max_id = 0
      else:
        max_id = self.address2id.max()
      input_ids = torch.arange(start=max_id, end=n_data, device=input.device)
    self.address2id[write_address] = input_ids
    self.is_empty[write_address] = 0

    # update number of stored items in each division
    self.div_size[unique_labels] += counts

    if self.verbose > 1:
      print(f"{n_data} new items added")

    return input_ids

  def train(self, input, force_retrain=False):
    """
      input: torch.Tensor, shape : [d_vector, n_data], dtype : float32
      force_retrain: bool, default : False
    """
    if self._is_trained == True and not force_retrain:
      print("Module is already trained")
      return
    d_vector, n_data = input.shape
    assert d_vector == self.d_vector
    if self.distance == "cosine":
      input = self.normalize(input)

    if self.verbose > 0:
      print("Start training coarse quantizer...")
    self.coarse_q.fit(input)

    if self.verbose > 0:
      print("Start training product quantizer...")
    self.product_q.train_codebook(input)

    self._is_trained.data = torch.tensor(True)
    if self.verbose > 0:
      print("Successfully trained!")
  
  def encode(self, input):
    """
      input: torch.Tensor, shape : [d_vector, n_data], dtype : float32
      return: torch.Tensor, shape : [n_subvectors, n_data], dtype : uint8
    """
    assert self._is_trained == True, "Module is not trained"
    assert input.shape[0] == self.d_vector
    if self.distance == "cosine":
      input = self.normalize(input)
    code = self.product_q.encode(input)
    return code

  def decode(self, code):
    """
      code: torch.Tensor, shape : [n_subvectors, n_data], dtype : uint8
      return: torch.Tensor, shape : [d_vector, n_data], dtype : float32
    """
    assert self._is_trained == True, "Module is not trained"
    assert code.shape[0] == self.n_subvectors
    recon = self.product_q.decode(code)
    return recon
    
  def topk(self, query, k, mode=2):
    """
      query: torch.Tensor, shape : [d_vector, n_query], dtype : float32
      k: int
    """
    assert self._is_trained == True, "module is not trained"
    d_vector, n_query = query.shape
    assert d_vector == self.d_vector
    assert 0 < k <= self.tot_size

    centroids = self.coarse_q.centroids
    sims = self.coarse_q.euc_sim(query, centroids)
    _, topk_labels = sims.topk(k=self.n_probe, dim=1)
    # topk_labels = torch.sort(topk_labels, dim=1)[0]
    div_start = self.div_start[topk_labels].int()
    if mode == 1: div_size = self.div_capacity[topk_labels].int()
    elif mode == 2: div_size = self.div_size[topk_labels].int()

    query = query.reshape(self.n_subvectors, self.d_subvector, n_query)
    codebook = self.product_q.codebook #[n_subvectors, d_subvector, n_pq_clusters]
    # if self.distance in ["cosine", "inner"]:
    #   precomputed = self.product_q.kmeans.cos_sim(query, codebook, normalize=False)
    # elif self.distance in ["euclidean"]:
    #   precomputed = self.product_q.kmeans.euc_sim(query, codebook)
    precomputed = self.product_q.kmeans.sim(query, codebook, normalize=False)

    topkv, topk_address = self._topk_fn(
      k=k,
      data=self.storage,
      precomputed=precomputed,
      is_empty=self.is_empty,
      div_start=div_start,
      div_size=div_size,
    )
    topki = self.get_id_of_address(topk_address.long())
    return (topkv, topki)


from .ComputeProductCUDA import ComputeProductCUDA
from .GetAddressOfIDCUDA import GetAddressOfIDCUDA
from .GetDivOfAddressCUDA import GetDivOfAddressCUDA
from .GetIOACUDA import GetIOACUDA
from .GetWriteAddressCUDA import GetWriteAddressCUDA
from .CustomKernel import CustomKernel
from .CustomKernel import Stream


import torch
import cupy as cp
import numpy as np
import math
from torchpq.kernels.CustomKernel import CustomKernel, Stream
from torchpq.util import get_absolute_path

class ComputeCentroidsCUDA(CustomKernel):
  def __init__(
      self,
      de=16,
      dk=16,
      sm_size=48*256*4,
    ):
    self.de = de
    self.dk = dk
    assert dk * (de + 1) * 4 <= sm_size
    self.tpb = 256
    self.sm_size = sm_size
    self._use_torch_in_cupy_malloc()
    self.stream = Stream(torch.cuda.current_stream().cuda_stream)

    with open(get_absolute_path("kmeans", "kernels", "ComputeCentroidsKernel.cu"), "r") as f:
      self.kernel = f.read()

    kernel = (self.kernel
      .replace("_DE_", str(de))
      .replace("_DK_", str(dk))
      .replace("_TPB_", str(self.tpb))
      .replace("_NITERS_", str(math.ceil(dk / self.tpb)))
    )

    self.fn = cp.RawKernel(
      kernel,
      'compute_centroids',
      # options=('--maxrregcount=255',),
      # backend='nvcc',
    )

    self.fn.max_dynamic_shared_size_bytes = sm_size
    # print(self.fn.attributes)

  def __call__(self, data, labels, k, centroids=None):
    """
      data: shape = [m, d_vector, n_data]   ###preferably [m, d, n_data]
      labels: shape = [m, n_data]
      centroids: shape = [m, n_clusters, d_vector]
    """
    m, d_vector, n_data = data.shape
    assert labels.shape[1] == n_data
    if centroids is not None:
      assert centroids.shape[0] == d_vector

      data = data.contiguous()
      labels = labels.contiguous()

    # data = data.transpose(1, 2).contiguous()

    centroids = torch.zeros(m, d_vector, k, device="cuda:0", dtype=torch.float32)
    threads_per_block = (self.tpb,)
    blocks_per_grid = (m, math.ceil(k / self.dk) , math.ceil(d_vector / self.de))
    # print("Blocks per grid", blocks_per_grid)

    torch.cuda.synchronize()
    self.fn(
      grid=blocks_per_grid,
      block=threads_per_block,
      shared_mem = self.dk * (self.de + 1) * 4,
      args=[
        data.data_ptr(),
        labels.data_ptr(),
        centroids.data_ptr(),
        m, # n_subquantizers / n_subvectors
        n_data,
        d_vector,
        k # n_clusters
        ],
      stream=self.stream
    )
    torch.cuda.synchronize()
    return centroids


import torch
import torch.nn as nn
import numpy as np

from .kernels import GetAddressOfIDCUDA
from .kernels import GetDivOfAddressCUDA
from .kernels import GetIOACUDA
from .kernels import GetWriteAddressCUDA

class IVFPQBase(nn.Module):
  def __init__(
    self,
    code_size,
    n_cq_clusters,
    blocksize,
    verbose,
    distance,
    device
    ):
    """
    this is a base class for IVFPQ variants
    for more details, you can search 'Product Quantization for Nearest Neighbor Search'

    Parameters:
      d_vector: int
        dimentionality of vectors to be quantized.

      code_size: int, default : 8
        byte size of quantized code, must to be a multiple of 4
        maximum possible code_size depends on GPU architecture
          GPU Architecture:   fp32  fp16 
          Ampere (GA100)      163   326  
          Turing (TU102 etc.) 64    128  
          Volta (GV100)       96    192  
          Pascal and before   48    96   
      
      n_cq_clusters: int, default : 128
        number coarse quantizer clusters
        recommended value is between 4*sqrt(n_data) ~ 16*sqrt(n_data)

      blocksize: int, default : 64
        number of vectors that can be asigned to each cluster of coarse_quantizer initially
        can be expanded using .expand methods, .add method will automatically cal .expand if necessary
        recommended value is (n_data_to_be_stored / n_cq_clusters)

      verbose: int, default : 0
        verbosity

      distance: str, default : 'euclidean'
        type of distance metric
        can be one of: ['euclidean', 'cosine', 'inner']

      device: str, default : 'cuda:0'
        which device to use
        can be one of: ['cpu', 'cuda', 'cuda:X'] *X represents cuda device index
      
    """
    super(IVFPQBase, self).__init__()
    n_cs = 4
    self.n_cs = n_cs
    self.code_size = code_size
    assert code_size % self.n_cs == 0, f"code_size needs to be a multiple of {n_cs}"
    self.n_cq_clusters = n_cq_clusters
    self.blocksize = blocksize
    self.verbose = verbose
    if distance.lower() in ["euclidean", "l2"]:
      distance = "euclidean"
    elif distance.lower() in ["dot", "inner"]:
      distance = "inner"
    elif distance.lower() in ["cosine"]:
      distance = "cosine"
    self.distance = distance
      
    self.register_buffer("_is_trained", torch.tensor(False))
    self.device = device

    storage = torch.zeros(
      code_size // n_cs,
      blocksize * n_cq_clusters,
      n_cs,
      device=self.device,
      dtype=torch.uint8,
    )
    address2id = torch.ones(
      blocksize * n_cq_clusters,
      device=self.device,
      dtype=torch.long,
    ) * -1
    is_empty = torch.ones(
      blocksize * n_cq_clusters,
      device=self.device,
      dtype=torch.uint8
    )
    div_start = torch.arange(
      n_cq_clusters,
      device=self.device
    ) * blocksize
    div_capacity = torch.ones(
      n_cq_clusters,
      device=self.device,
      dtype=torch.long,
    ) * blocksize
    div_size = torch.zeros(
      n_cq_clusters,
      device=self.device,
      dtype=torch.long
    )
    self.register_buffer("storage", storage)
    self.register_buffer("div_start", div_start)
    self.register_buffer("div_capacity",div_capacity)
    self.register_buffer("div_size", div_size)
    self.register_buffer("address2id", address2id)
    self.register_buffer("is_empty", is_empty)

    self._get_address_of_id_cuda = GetAddressOfIDCUDA(
      tpb=256,
    )
    self._get_div_of_address_cuda = GetDivOfAddressCUDA(
      ta=1,
      tpb=256,
    )
    self._get_ioa_cuda = GetIOACUDA(
      tpb=256,
    )
    self._get_write_address_cuda = GetWriteAddressCUDA(
      tpb=256,
    )

  def __getitem__(self, ids):
    """
      ids: int, or an iterable of integers
    """
    if type(ids) is not torch.Tensor:
      ids = torch.LongTensor(ids)
    if len(ids.shape) == 0:
      ids = ids[None]
    elif len(ids.shape) > 1:
      raise ValueError("only 1D array/tensors are supported")
    return self.get_data_of_id(ids)

  def __repr__(self):
    def get_bytesize_str(bytesize):
      if bytesize < 1024*1:
        bytesize = f"{bytesize} B"
      elif bytesize < 1024**2:
        bytesize = f"{round(bytesize / 1024, 2)} KB"
      elif bytesize < 1024**3:
        bytesize = f"{round(bytesize / 1024**2, 2)} MB"
      elif bytesize < 1024**4:
        bytesize = f"{round(bytesize / 1024**3, 2)} GB"
      elif bytesize < 1024**5:
        bytesize = f"{round(bytesize / 1024**4, 2)} TB"
      return bytesize

    bytesize = self.bytesize
    cpu_bytesize = get_bytesize_str(bytesize[0])
    gpu_bytesize = get_bytesize_str(bytesize[1])

    txt =  f"""
    {type(self).__name__}(
      number of stored items = {self.tot_size},
      total capacity of storage = {self.tot_capacity},
      byte size on CPU = {cpu_bytesize},
      byte size on GPU = {gpu_bytesize},
      code size = {self.code_size} B,
      is trained = {self._is_trained.item()}
    )
    """
    txt = '\n'.join([i[4:] for i in txt.split('\n')])
    return txt

  def __len__(self):
    return self.tot_size

  @staticmethod
  def get_cc():
    if torch.cuda.is_available():
      device_id = torch.cuda.current_device()
      gpu_properties = torch.cuda.get_device_properties(device_id)
      result = (gpu_properties.major, gpu_properties.minor)
    else:
      result = (10, 0)
    return result
  
  @staticmethod
  def normalize(a, dim=-1):
    """
      normalize given tensor along given dimention
    """
    a_norm = a.norm(dim=dim, keepdim=True) + 1e-9
    return a / a_norm

  @property
  def tot_size(self):
    """
      total number of items stored
    """
    return self.div_size.sum().item()

  @property
  def tot_capacity(self):
    """
      total capacity of storage
    """
    return self.div_capacity.sum().item()

  @property
  def bytesize(self):
    """
      the bytesize of cpu and gpu memory that are being used by module
      returns: (cpu_bytesize, gpu_bytesize)
    """
    cpu_bytesize = 0
    gpu_bytesize = 0
    for p in self.state_dict().values():
      if type(p) is torch.Tensor:
        if p.device.type == "cpu":
          cpu_bytesize += p.numel() * p.element_size()
        elif p.device.type == "cuda":
          gpu_bytesize += p.numel() * p.element_size()
    return (cpu_bytesize, gpu_bytesize)

  def _get_address_of_id_old(self, ids):
    batch_size = ids.shape[0]
    address = torch.zeros(batch_size, device=self.device, dtype=torch.long)
    for i in range(batch_size):
      id = ids[i]
      adr = torch.nonzero(self.address2id == id)
      if adr.shape[0] > 0:
        address[i] = adr[0, 0]
      else:
        address[i] = -1
    return address

  def get_address_of_id(self, ids):
    """
      ids: torch.Tensor, shape : [batch_size], dtype : int64
      if there are multiple matching ids, only the address of the first one will be returned
      if there are no matching ids, the address is -1
    """
    if self.address2id.device.type == "cuda":
      return self._get_address_of_id_cuda(self.address2id, ids)
    elif self.address2id.device.type == "cpu":
      return self._get_address_of_id_old(ids)

  def get_id_of_address(self, address):
    """
      address: torch.Tensor, shape : [batch_size], dtype : int64
    """
    return self.address2id[address]

  def _get_div_of_address_old(self, address):
    assert address.max() < self.tot_capacity
    batch_size = address.shape[0]
    div_start = self.div_start
    div_end = div_start + self.div_capacity
    # for adr in address:
    #   mask1 = div_start <= adr
    #   mask2 = div_ends > adr
    #   mask = mask1 & mask2
    #   div = torch.nonzero(mask)
    mask1 = div_start[None, ] <= address[:, None]
    mask2 = div_end[None, ] > address[:, None] # [batch_size, n_cq_clusters]
    mask = mask1 & mask2
    not_found = mask.sum(dim=1) == 0
    mask[not_found, 0] = True
    divs = torch.nonzero(mask)
    divs[not_found, 1] = -1
    return divs[:, 1]

  def get_div_of_address(self, address):
    """
      address: torch.Tensor, shape : [n_address], dtype : int64
      if address is not in range [0, tot_capacity), its div will be -1
    """
    if self.div_start.device.type == "cuda":
      div_start = self.div_start
      div_end = div_start + self.div_capacity
      return self._get_div_of_address_cuda(address, div_start, div_end)
    elif self.div_start.device.type == "cpu":
      return self._get_div_of_address_old(address)

  def get_data_of_address(self, address):
    """
      address: torch.Tensor, shape : [n_address], dtype : int64
      returns: torch.Tensor, shape : [code_size, n_address], dtype : uint8
      address must be in range [0, tot_capacity)
    """
    n_address = address.shape[0]
    assert address.dtype == torch.long
    mask = address < 0
    data = self.storage[:, address] #[code_size//n_cs, n_address, n_cs]
    data = data.transpose(1,2).reshape(self.code_size, n_address)
    if mask.sum() > 0:
      data[:, mask] = 0
    return data

  def get_data_of_id(self, ids):
    """
      ids: torch.Tensor, shape : [n_ids], dtype : int64
      returns: torch.Tensor, [code_size, n_ids]
      if an id doesn't exist, its retrieved data will be zeros.
    """
    address = self.get_address_of_id(ids)
    data = self.get_data_of_address(address)
    return data

  def set_data_of_address(self, data, address):
    """
      data: torch.Tensor, shape : [code_size // n_cs, n_data, n_cs], dtype : uint8
      address: torch.Tensor, shape : [n_data], dtype : int64
    """
    self.storage[:, address] = data

  def set_data_of_id(self, data, ids):
    """
      data: torch.Tensor, shape : [code_size // n_cs, n_data, n_cs], dtype : uint8
      ids: torch.Tensor, shape : [n_data], dtype : int64
    """
    address = self.get_address_of_id(ids)
    set_data_of_address(data, address)

  def _get_ioa_old(self, labels, unique_labels=None):
    if unique_labels is None:
      unique_labels = torch.unique(labels) #[n_unique_clusters]
    expanded_labels = labels[:, None].expand(-1, unique_labels.shape[0]) #[n_data, n_unique_clusters]
    mask = expanded_labels == unique_labels[None, :] #[n_data, n_unique_clusters]
    mcs = mask.cumsum(dim=0)
    mcs[[~mask]] = 0
    ioa = mcs.sum(dim=1) - 1

  def get_ioa(self, labels, unique_labels=None):
    if labels.device.type == "cuda":
      return self._get_ioa_cuda(labels, unique_labels)
    elif labels.device.type == "cpu":
      return self._get_ioa_old(labels, unique_labels)

  def _get_write_address_old(self, empty_adr, div_of_empty_adr, labels, ioa):
    write_address = torch.zeros_like(labels)
    for i in range(n_labels):
      div_mask = div_of_empty_adr == labels[i]
      write_adr = empty_adr[div_mask]
      write_adr = write_adr[ioa[i] ]
      write_address[i] = write_adr
    return write_address

  def get_write_address(self, empty_adr, labels, div_of_empty_adr=None, ioa=None):
    if div_of_empty_adr is None:
      div_of_empty_adr = self.get_div_of_address(empty_adr)
    if ioa is None:
      ioa = self.get_ioa(labels)

    if labels.device.type == "cuda":
      return self._get_write_address_cuda(empty_adr, div_of_empty_adr, labels, ioa)
    elif labels.device.type == "cpu":
      return self._get_write_address_old(empty_adr, div_of_empty_adr, labels, ioa)

  def expand(self, clusters):
    """
        clusters: an iterable of integers, len(clusters) <= n_cq_clusters
    """
    if self.verbose > 1:
      print(f"Expanding {len(clusters)} clusters")
    tot = 0
    storage = self.storage
    address2id = self.address2id
    is_empty = self.is_empty
    del self.storage, self.address2id, self.is_empty
    for cluster_index in clusters:
      div_start = self.div_start[cluster_index]
      div_cap = self.div_capacity[cluster_index].item()
      div_end = div_start + div_cap
      new_block = torch.zeros(
        self.code_size // self.n_cs,
        div_cap,
        self.n_cs, device=self.device,
        dtype=torch.uint8,
      )
      storage = torch.cat([
        storage[:, :div_end],
        new_block,
        storage[:, div_end:]
      ], dim=1)

      new_a2i = torch.ones(div_cap, device=self.device, dtype=address2id.dtype) * -1
      address2id = torch.cat([
        address2id[:div_end],
        new_a2i,
        address2id[div_end:]
      ], dim=0)

      new_is_empty = torch.ones(div_cap, device=self.device, dtype=is_empty.dtype)
      is_empty = torch.cat([
        is_empty[:div_end],
        new_is_empty,
        is_empty[div_end:]
      ], dim=0)

      self.div_capacity[cluster_index] += div_cap
      arange = torch.arange(start=cluster_index+1, end=self.n_cq_clusters)
      self.div_start[arange] += div_cap
      # self.div_start[:] = self.div_capacity.cumsum(dim=0)
      tot += div_cap
    self.register_buffer("storage", storage)
    self.register_buffer("address2id", address2id)
    self.register_buffer("is_empty", is_empty)
    if self.verbose > 1:
      print(f"Storage capacity is increased by {tot}, total capacity: {self.storage.shape[1]}")
  
  def add(self, input, input_ids=None):
    """
      has to be overrided in child class
    """
    raise NotImplementedError("method needs to be overrided in child class")

  def remove(self, remove_ids):
    """
      remove_ids: torch.Tensor, shape : [n_ids], dtype : int64
    """
    assert self._is_trained == True, "Module is not trained"
    n_ids = remove_ids.shape[0]
    address = self.get_address_of_id(remove_ids)
    address = address[address >= 0] #filter out -1 values
    divs = self.get_div_of_address(address)

    self.is_empty[address] = 1
    self.address2id[address] = -1

    unique_divs, counts = divs.unique(return_counts=True)
    self.div_size[unique_divs] -= counts
    if self.verbose > 1:
      print(f"{len(address)} items have been successfully removed, {n_ids - len(address)} items are ignored.")

  def train(self, input, force_retrain=False):
    """
      has to be overrided in child class
    """
    raise NotImplementedError("method needs to be overrided in child class")
  
  def encode(self, input):
    """
      has to be overrided in child class
    """
    raise NotImplementedError("method needs to be overrided in child class")

  def decode(self, code):
    """
      has to be overrided in child class
    """
    raise NotImplementedError("method needs to be overrided in child class")
    
  def topk(self, query, k, mode=1):
    """
      has to be overrided in child class
    """
    raise NotImplementedError("method needs to be overrided in child class")
    


import torch
import torch.nn as nn
import numpy as np
import math

from .kernels import MaxSimCUDA
from .kernels import ComputeCentroidsCUDA

class MultiKMeans(nn.Module):
  """
  Run multiple independent K-means algorithms in parallel.
  Parameters:
    n_clusters: int, 
      Number of clusters

    max_iter: int, default: 100
      Maximum number of iterations

    tol: float, default: 0.0001
      Tolerance

    n_redo: int, default: 1
      Number of time k-means will be run with differently intialized centroids.
      the centroids with the lowest inertia will be selected as a final result.

    init_mode: {'random', 'kmeans++'}, default: 'random'
      Initialization method
      'random': randomly chose initial centroids from input data
      'kmeans++': use k-means++ algorithm to initialize centroids, slow when n_cluster is large, but converges faster)
    
    verbose: int, default: 0
      Verbosity

    distance: {'euclidean', 'cosine', 'manhattan'}, default: 'euclidean'
      Type of distance metric
      note: manhattan or L1 distance is only supported on GPU
      
  Attributes:
    centroids: torch.Tensor, shape: [d_vector, n_clusters]
      cluster centroids
  """
  def __init__(
    self,
    n_clusters,
    n_redo=1,
    max_iter=100,
    tol=1e-4,
    distance="euclidean",
    init_mode="random",
    verbose=0,
    sm_size=48*256*4,
   ):
    super(MultiKMeans, self).__init__()
    self.n_redo = n_redo
    self.n_clusters = n_clusters
    self.max_iter = max_iter
    self.tol = tol
    self.verbose = verbose
    self.distance = distance
    self.init_mode = init_mode
    self.sm_size = sm_size
    self.arange = None
    if n_clusters < 4096:
      dk = n_clusters
    else:
      dk = 4096
    de = 1

    self.register_buffer("centroids", None)
    
    if torch.cuda.is_available():
      self.compute_centroids_cuda = ComputeCentroidsCUDA(
        de=de,
        dk=dk,
        sm_size=sm_size,
      )

      if distance in ["euclidean", "manhattan"]:
        distance = distance
      elif distance in ["cosine"]:
        distance = "inner"
      self.max_sim_cuda = MaxSimCUDA(
        distance=distance,
      )

  @staticmethod
  def remaining_memory(device):
    """
      Get remaining memory of GPU in bytes
    """
    # torch.cuda.synchronize()
    if device.type == "cpu":
      remaining = 32 * 1024 ** 3 # just a random large number
    elif device.type == "cuda":
      torch.cuda.empty_cache()
      total_memory = torch.cuda.get_device_properties(0).total_memory
      remaining = total_memory - torch.cuda.memory_reserved()
      # remaining = total_memory - torch.cuda.memory_allocated()
    return remaining

  @staticmethod
  def does_it_fit(size, device="cpu", dtype=torch.float):
    try:
      torch.empty(size, device=device, dtype=dtype)
    except:
      return False
    else:
      return True

  @staticmethod
  def calculate_error(a, b):
    """
      Compute L2 error between 'a' and 'b'
    """
    diff = a - b
    diff.pow_(2)
    return diff.sum()

  @staticmethod
  def cos_sim(a, b, normalize=True, inplace=False):
    """
      Compute batched cosine similarity between 'a' and 'b'

      a: torch.Tensor, shape : [n_kmeans, d_vector, m]
      b: torch.Tensor, shape : [n_kmeans, d_vector, n]
      normalize: bool, default : True
        if True, a and b will be normalized to norm=1
      inplace: bool, default : False
      returns: torch.Tensor, shape : [l, m, n]
    """
    if normalize:
      a_norm = a.norm(dim=-2, keepdim=True) + 1e-8 #[l, m] <l*(m*4)>
      b_norm = b.norm(dim=-2, keepdim=True) + 1e-8 #[l, n] <l*(m*4 + n*4)>
      if inplace:
        # memory consump: m + n + (m * n)
        a.div_(a_norm)
        b.div_(b_norm)
      else:
        # memory consum: m + n + (m * n) + m*d + n*d
        a = a / a_norm #[l, d_vector, m], l*(<m*4 + n*4> + <m*d*4>)
        b = b / b_norm #[l, d_vector, n], l*(<m*4 + n*4> + <(m+n)*d*4>)
    prod = a.transpose(-2, -1) @ b #[l, m, n], <m*n*4 + m*4 + n*4> + <(m+n)*d*4>
    if inplace and normalize:
      a.mul_(a_norm)
      b.mul_(b_norm)
    return prod

  @staticmethod
  def euc_sim(a, b, inplace=False):
    """
      Compute batched negative squared euclidean distance between 'a' and 'b'
      a: torch.Tensor, shape : [l, d_vector, m]
      b: torch.Tensor, shape : [l, d_vector, n]
      inplace: bool, default : False
      returns: torch.Tensor, shape : [l, m, n]
    """
    # peak mem uwage: m*n*4 + max(m,n)*4 + inplace ? 0: (m+n)*d*4
    y = a.transpose(-2, -1) @ b # [m, n] <m*n*4>
    y.mul_(2)
    if inplace:
      a.pow_(2)
      b.pow_(2)
    else:
      a = a ** 2 #[m, d], <m*n*4 + m*d*4>
      b = b ** 2 #[n, d], <m*n*4 + n*d*4 + m*d*4>
    a2 = a.sum(dim=-2)[..., :, None] #? [m], <m*n*4 + m*4> + <n*d*4 + m*d*4>
    y.sub_(a2)
    del a2
    b2 = b.sum(dim=-2)[..., None, :] #[n], <m*n*4 + n*4> + <n*d*4 + m*d*4>
    y.sub_(b2)
    if inplace:
      a.sqrt_()
      b.sqrt_()
    return y
  
  def sim(self, a, b, inplace=False, normalize=True):
    """
      Compute batched similarity between 'a' and 'b', the type of distance metric is specified in __init__ method
      a: torch.Tensor, shape : [l, d, m]
      b: torch.Tensor, shape : [l, d, n]
      returns: torch.Tensor, shape : [l, m, n]
    """
    if self.distance == "euclidean":
      return self.euc_sim(a, b, inplace=inplace)
    elif self.distance == "cosine":
      return self.cos_sim(a, b, inplace=inplace, normalize=normalize)

  ### Need more testing:
  def kmeanspp(self, data):
    """
      Initialize centroids with k-means++ algorithm
      data: torch.Tensor, shape : [l, d_vector, n_data]
      returns: torch.Tensor, shape : [l, d_vector, n_clusters]
    """
    l, d_vector, n_data = data.shape
    if self.distance == "cosine":
      data_norm = data.norm(dim=-2, keepdim=True) + 1e-8
      data.div_(data_norm)
    centroids = torch.zeros(l, d_vector, self.n_clusters, device=data.device, dtype=data.dtype)
    #Select initial centroid
    centroids[:, :, 0] = data[:, :, np.random.randint(n_data)]
    for i in range(1, self.n_clusters):
      current_centroids = centroids[:, :, :i].contiguous()
      if data.device.type == "cpu":
        sims = self.sim(data, current_centroids ) #[l,m,n]
        max_sims_v, max_sims_i = sims.max(dim=-1) #[l,m]
      elif data.device.type == "cuda":
        max_sims_v, max_sims_i = self.max_sim_cuda(data, current_centroids, dim=2, distance="tn")
      index = max_sims_v.argmin(dim=-1) #[l]
      arange = torch.arange(l, device=device)
      new_centroid = data[arange, :, index] #[l, d_vector]
      centroids[:, :, i] = new_centroid
    if self.distance == "cosine":
      data.mul_(data_norm)
    return centroids

  def initialize_centroids(self, data):
    """
      Initialize centroids with init_method specified in __init__
      data: torch.Tensor, shape : [l, d_vector, n_data]
      return: torch.Tensor, shape: [l, d_vector, n_clusters]
    """
    l, d_vector, n_data = data.shape
    if self.init_mode == "random":
      random_index = np.random.choice(
        n_data,
        size=[self.n_clusters],
        replace=False
      )
      centroids = data[:, :, random_index]
      if self.verbose >= 1:
        print("centroids are randomly initialized.")

    elif self.init_mode == "kmeans++":
      centroids = self.kmeanspp(data)
      if self.verbose >= 1:
        print("kmeans++ initialization is done!")
    return centroids

  def get_labels(self, data, centroids):
    """
      data: torch.Tensor, shape : [l, d_vector, n_data]
      centroids: torch.Tensor, shape : [l, d_vector, n_clusters]
      return: torch.Tensor, shape : [l, n_data], dtype: long
    """
    #memory requirement: 

    l, d, m = data.shape
    l, d, n = centroids.shape

    remaining = self.remaining_memory(data.device)# - 1024*3

    if self.distance == "euclidean":
      required = l*(m*n + max(m, n) + m*d + n*d) * data.element_size()
    elif self.distance == "cosine":
      required = l*((m*n) + (m+n)*(d+1)) * data.element_size()
    if remaining >= required:
      sims = self.sim(data, centroids, inplace=False) #[l, m, n]
      maxsims, labels = sims.max(dim=-1) #[l, m]
      return (maxsims, labels)
    else:
      if data.device.type == "cuda":
        if self.distance == "cosine":
          d_norm = data.norm(dim=-2, keepdim=True) + 1e-8
          c_norm = centroids.norm(dim=-2, keepdim=True) + 1e-8
          data.div_(d_norm)
          centroids.div_(c_norm)
        maxsims, labels = self.max_sim_cuda(data, centroids, dim=2, distance="tn")
        if self.distance == "cosine":
          data.mul_(d_norm)
          centroids.mul_(c_norm)
      elif data.device.type == "cpu":
        ## doing in seperate chunks
        n_partitions = 1
        for i in range(16):
          sub_m = math.ceil(m / n_partitions)
          if self.distance == "euclidean":
            required = l*(sub_m*n + max(sub_m, n)) * data.element_size() + m*8 # +sub_m*d*4
          elif self.distance == "cosine":
            required = l*(sub_m*n + sub_m+n) * data.element_size() + m*8# +sub_m*d*4
          # print("required, remaining, n_p", required / 1024**3, remaining / 1024**3, n_partitions)
          if self.does_it_fit(required // 4, device=data.device, dtype=torch.float):
            break
          n_partitions *= 2

        maxsims = torch.zeros(l, m, device=data.device, dtype=torch.float)
        labels = torch.zeros(l, m, device=data.device, dtype=torch.long)
        for i in range(n_partitions):
          start = i*sub_m
          if i == n_partitions - 1:
            end = m - 1
          else:
            end = (i+1)*sub_m
          sub_data = torch.narrow(data, dim=-1, start=start, length=end-start) #[l, d, sub_m]
          # sub_data = data[:, start:end] #[l, d, sub_m]
          sub_sims = self.sim(sub_data, centroids, inplace=True) #[l, sub_m, n]
          del sub_data
          sub_maxsims, sub_labels = sub_sims.max(dim=-1) #[l, sub_m]
          del sub_sims
          labels[:, start:end] = sub_labels
          maxsims[:, start:end] = sub_maxsims
          del sub_labels
      return (maxsims, labels)

  def compute_centroids_loop(self, data, labels):
    """
      data: torch.Tensor, shape : [l, d_vector, n_data]
      labels: torch.Tensor, shape : [l, n_data]
      return: torch.Tensor, shape : [l, d_vector, n_clusters]
    """
    ### Naive method with loop
    l, d, m = data.shape
    centroids = torch.zeros(l, d, self.n_clusters, device=data.device, dtype=data.dtype)
    for j in range(l):
      unique_labels, counts = labels[j].unique(return_counts=True)
      for i, count in zip(unique_labels, counts):
        centroids[j, :, i] = data[j, :, labels[j]==i].sum(dim=1) / count
    return centroids

  def compute_centroids(self, data, labels):
    """
      data: torch.Tensor, shape : [l, d_vector, n_data]
      labels: torch.Tensor, shape : [l, n_data]
      return: torch.Tensor, shape: [l, d_vector, n_clusters]
    """
    if self.n_clusters <= 256 or data.device == torch.device("cpu"):
      centroids = self.compute_centroids_loop(data, labels)
    else:
      centroids = self.compute_centroids_cuda(data, labels, k=self.n_clusters)
    return centroids

  def _compute_centroids_hungry(self, data, labels):
    ### Memory hungry method
    # expanded_labels = labels[None].expand(self.n_clusters, -1) #[k, n], k=n_clusters <>
    if self.arange is None\
    or self.arange.dtype != data.dtype\
    or self.arange.device != data.device:
      self.arange = torch.arange(self.n_clusters, device=data.device) #[k] <k*8>
    
    mask = labels[None, :] == self.arange[:, None] #[k, n] <k*n*1 + k*8>
    mask_sum = mask.sum(dim=-1) #[k] <k*n*1 + k*12>
    mask = mask.float() # <k*n*5 + k*12> LARGEST MEMORY USE!!!
    centroids = mask @ data # <k*n*4 + k*12 + k*d*4>
    del mask

    centroids.div_(mask_sum[..., :, None]) # <k*d*4 + k*12>
    del mask_sum
    nan_mask = centroids!=centroids #[k, d] # <k*d*8>
    centroids[nan_mask] = 0 # remove NaNs

    return centroids

  def fit(self, data):
    """
      Perform K-means clustering, and return final labels
      data: torch.Tensor, shape : [l, d_vector, n_data]
      return: torch.Tensor, shape : [l, n_data], dtype: long
    """
    assert data.is_contiguous(), "use .contiguous()"

    best_centroids = None
    best_error = 1e32
    best_labels = None
    best_inertia = 1e32
    tm = time()
    for i in range(self.n_redo):
      tm_i = time()
      centroids = self.initialize_centroids(data)
      for j in range(self.max_iter):
        # 1 iteration of clustering
        maxsims, labels = self.get_labels(data, centroids) #top1 search
        new_centroids = self.compute_centroids(data, labels)
        error = self.calculate_error(centroids, new_centroids)
        centroids = new_centroids
        inertia = torch.sum(1-maxsims)
        if self.verbose >= 3:
          print(f"----iteration {j} of {i}th redo, error={error.item()}, inertia={inertia.item()}")
        if error <= self.tol:
          break

      if inertia < best_inertia:
        best_centroids = centroids
        best_error = error
        best_labels = labels
        best_inertia = inertia
      if self.verbose >= 2:
        print(f"--{i}th redo finished, error: {error.item()}, inertia: {inertia.item()}time spent:{round(time()-tm_i, 4)} sec")

    self.register_buffer("centroids", best_centroids)
    if self.verbose >= 1:
      print(f"finished {self.n_redo} redos in {round(time()-tm, 4)} sec, final_inertia: {best_inertia}")
    return best_labels

  def predict(self, query):
    """
      Predict closest cluster center each sample in query belongs to.
      query: torch.Tensor, shape : [l, n_query, d_vector]
      return: torch.Tensor, shape : [l, n_query]
    """
    assert self.centroids is not None, "kmeans is not trained"
    _, labels = self.get_labels(query, self.centroids)
    return labels

  def topk(self, query, k):
    """
      Predict the top-k closest cluster centers of each sample in query
      query: torch.Tensor, shape : [l, n_query, d_vector]
      k: int, should be in range [1, n_centroids]
    """
    assert self.centroids is not None, "kmeans is not trained"
    assert k <= self.n_centroids, "k is too large"
    sims = self.sim(query, self.centroids) #[l, n_query, n_clusters]
    topkv, topki = sims.topk(dim=-1, k=k) #[l, n_query, k]
    return (topkv, topki)


import torch
import torch.nn as nn

from .kmeans import MultiKMeans

class PQ(nn.Module):
  def __init__(
      self,
      d_vector,
      n_subvectors=8,
      n_clusters=256,
      distance="euclidean",
      verbose=0
    ):
    super(PQ, self).__init__()
    self.n_subvectors = n_subvectors
    self.n_clusters = n_clusters
    self.d_vector = d_vector
    self.verbose = verbose
    assert d_vector % n_subvectors == 0

    self.d_subvector = d_vector // n_subvectors
    self.distance = distance
    
    self.kmeans = MultiKMeans(
      n_clusters = n_clusters,
      distance = distance,
      max_iter = 25,
      verbose = verbose,
    )

  @property
  def codebook(self):
    """
      return : torch.Tensor, 
        shape :[n_subvectors, d_subvectors, n_clusters]
    """
    return self.kmeans.centroids

  def train_codebook(self, x):
    """
      x: torch.Tensor, shape : [d_vector, n_data], dtype : float32
    """
    d_vector, n_data = x.shape
    assert d_vector == self.d_vector
    x = x.reshape(self.n_subvectors, self.d_subvector, n_data)
    return self.kmeans.fit(x)

  def encode(self, x):
    """
      x: torch.Tensor, shape : [d_vector, n_data], dtype : float32
      return: torch.Tensor, shape : [n_subvectors, n_data], dtype : uint8
    """
    d_vector, n_data = x.shape
    assert d_vector == self.d_vector
    x = x.reshape(self.n_subvectors, self.d_subvector, n_data)
    _, labels = self.kmeans.get_labels(x, self.codebook) #[n_subvectors, n_data]
    return labels.byte()

  def decode(self, code):
    """
      code: torch.Tensor, shape : [n_subvectors, n_data], dtype : uint8
      return: torch.Tensor, shape : [d_vector, n_data], dtype : float32
    """
    n_subvectors, n_data = code.shape
    assert n_subvectors == self.n_subvectors
    arange = torch.arange(n_subvectors)[:, None].expand(-1, n_data)
    res = self.codebook[arange, :, code.long()]
    res = res.transpose(1, 2).reshape(self.d_vector, n_data)
    return res


import torch
import cupy as cp
import numpy as np
import math
from .CustomKernel import CustomKernel, Stream
from torchpq.util import get_absolute_path

class ComputeProductCUDA(CustomKernel):
  def __init__(
      self,
      m=8,
      k=256,
      n_cs=4,
      sm_size=48*256*4,
    ):
    self.m = m
    self.k = k
    self.tpb = 256
    self.n_cs = n_cs
    self.sm_size = sm_size
    self._use_torch_in_cupy_malloc()
    self.stream = Stream(torch.cuda.current_stream().cuda_stream)

    with open(get_absolute_path("kernels", "ComputeProductKernel.cu"), "r") as f:
      self.kernel = f.read()
    
    cb1 = [f"      float Bval{i} = Bsh[(i * _NCS_ + {i}) * _K_ + int(Avals.d{i}) ];" for i in range(n_cs)]
    cb2 = [f"      sum += Bval{i};" for i in range(n_cs)]
    codeblock = "\n".join(cb1) + "\n" + "\n".join(cb2)
    varnames = ", ".join([f"d{i}" for i in range(n_cs)])
    kernel = (self.kernel
      .replace("_CODEBLOCK_", codeblock)
      .replace("_VARNAMES_", varnames)
      .replace("_M_", str(m))
      .replace("_K_", str(k))
      .replace("_TPB_", str(self.tpb))
      .replace("_NCS_", str(n_cs))
    )
    # print(kernel.split('\n')[60:64])
    self.fn = cp.RawKernel(
      kernel,
      'compute_product',
      # options=('--maxrregcount=255',),
      # backend='nvcc',
    )

    self.fn.max_dynamic_shared_size_bytes = sm_size
    # print(self.fn.attributes)

  def __call__(self, data, precomputed, is_empty, div_start, div_size, max_out_size=None):
    """
      data: shape=[n_subvectors // n_cs, n_data, n_cs], dtype=uint8
      precomputed: shape=[n_subvectors, n_query, n_clusters], dtype=float32
      is_empty: shape=[n_data], dtype=uint8
      div_start: shape=[n_query, n_probe], dtype=int32
      div_size: shape=[n_query, n_probe], dtype=int32
    """
    n_data = data.shape[1]
    n_query = precomputed.shape[1]
    n_probe = div_start.shape[1]
    assert precomputed.shape[0] == self.m
    assert precomputed.shape[2] == self.k
    assert data.shape[0] == self.m // self.n_cs
    assert data.shape[2] == self.n_cs
    assert is_empty.shape[0] == n_data
    assert div_size.shape[1] == n_probe
    assert data.dtype == torch.uint8
    assert precomputed.dtype == torch.float32
    assert div_start.dtype == div_size.dtype == torch.int32
    assert is_empty.dtype == torch.uint8

    if max_out_size is None:
      max_out_size = div_size.sum(dim=1).max().item()
    values = torch.empty(n_query, max_out_size, device="cuda:0", dtype=torch.float32)
    values.fill_(float("-inf"))
    indices = torch.zeros(n_query, max_out_size, device="cuda:0", dtype=torch.int32) #?

    threads_per_block = (self.tpb,)
    blocks_per_grid = (n_query,)
    self.fn(
      grid=blocks_per_grid,
      block=threads_per_block,
      shared_mem = self.sm_size,
      args=[
        data.data_ptr(),
        precomputed.data_ptr(),
        is_empty.data_ptr(),
        div_start.data_ptr(),
        div_size.data_ptr(),
        values.data_ptr(),
        indices.data_ptr(),
        n_data, n_query, max_out_size, n_probe,
        ],
      stream=self.stream
    )
    return (values, indices)


from setuptools import setup, find_packages
setup(
  name = 'torchpq',
  packages = find_packages(),
  version = '0.1.3.9.4',
  license='MIT',
  description = 'Efficient implementations of Product Quantization and its variants',
  author = 'demoriarty', 
  author_email = 'sahbanjan@gmail.com',
  url = 'https://github.com/DeMoriarty/TorchPQ',
  download_url = 'https://github.com/DeMoriarty/TorchPQ/archive/v_01394.tar.gz',
  keywords = ['KMeans', 'K-means', 'ANN', 'pytorch','machine learning', 'pq', 'product quantization', 'IVFPQ', 'approximate nearest neighbors'],
  install_requires=[ 
    'numpy',
    'torch',
  ],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
  ],
  include_package_data = True,
)


from .kernels import ComputeProductCUDA

class IVFPQ_Topk:
  def __init__(self,
    n_subvectors,
    n_clusters,
    n_cs=4,
    ):
    assert torch.cuda.is_available()
    self.n_subvectors = n_subvectors
    self.n_clusters = n_clusters
    self.n_cs = n_cs
    self.sm_size = n_subvectors * 256 * 4

    self.compute_product = ComputeProductCUDA(
      m=n_subvectors,
      k=n_clusters,
      n_cs=n_cs,
      sm_size=self.sm_size
    )

  @staticmethod
  def remaining_memory():
    if torch.cuda.is_available():
      torch.cuda.empty_cache()
      total_memory = torch.cuda.get_device_properties(0).total_memory
      remaining = total_memory - torch.cuda.memory_reserved()
    else:
      remaining = 0
    return remaining

  def __call__(self, k, data, precomputed, is_empty, div_start, div_size):
    """
      k: dtype : int
      data: shape : [n_subvectors // n_cs, n_data, n_cs], dtype : uint8
      precomputed: shape : [n_subvectors, n_query, n_clusters], dtype : float32
      is_empty: shape : [n_data], dtype : uint8
      div_start: shape : [n_query, n_probe], dtype : int32
      div_size: shape : [n_query, n_probe], dtype : int32
    """
    max_out_size = div_size.sum(dim=1).max().item()
    n_subvectors, n_query, n_clusters = precomputed.shape
    n_probe = div_start.shape[1]


    final_v = torch.zeros(n_query, k, device="cuda:0", dtype=torch.float32)
    final_i = torch.zeros(n_query, k, device="cuda:0", dtype=torch.int32)
    remaining = self.remaining_memory()
    n_partitions = 1
    while True:
      if n_partitions > n_query:
        raise RuntimeError("No enough GPU memory")
      sub_n_query = math.ceil(n_query / n_partitions)
      required = sub_n_query * max_out_size * 2 * 4
      if n_partitions > 1:
        required += sub_n_query * n_subvectors * n_clusters * 4
        required += sub_n_query * n_probe * 2 * 4
      if required <= remaining:
        break
      n_partitions *= 2
    for i in range(n_partitions):
      start = i * sub_n_query
      end = (i+1) * sub_n_query
      if end > n_query:
        end = n_query
      if n_partitions > 1:
        sub_precomputed = precomputed[:, start:end].contiguous()
        sub_div_start = div_start[start:end].contiguous()
        sub_div_size = div_size[start:end].contiguous()
        sub_mos = sub_div_size.sum(dim=1).max().item()
      else:
        sub_precomputed = precomputed
        sub_div_start = div_start
        sub_div_size = div_size
        sub_mos = max_out_size

      sub_v, sub_i = self.compute_product(
        data = data,
        precomputed = sub_precomputed,
        is_empty = is_empty,
        div_start = sub_div_start,
        div_size = sub_div_size,
        max_out_size = sub_mos,  
      )
      del sub_precomputed
      sorted_v, sorted_i = torch.topk(sub_v, dim=-1, k=k)
      del sub_v
      final_v[start:end] = sorted_v
      del sorted_v
      final_i[start:end] = torch.gather(input=sub_i, index=sorted_i, dim=1)
      del sub_i, sorted_i

    ### TEST
    # def naive_pqd(data, distances, is_empty):
    #   o, n, q = data.shape
    #   m = o * q
    #   arange = torch.arange(m, device="cuda:0")
    #   data = data.transpose(0, 1).reshape(n,m)
    #   data = data[~is_empty ]
    #   result = distances[arange, :, data[:].long() ].sum(dim=1).t()
    #   return result

    return (final_v, final_i)